@@ -1,7 +1,7 @@
 /***
   This file is part of PaSystray
 
-  Copyright (C) 2011, 2012 Christoph Gysin
+  Copyright (C) 2011-2015  Christoph Gysin
 
   PaSystray is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
@@ -27,7 +27,7 @@
 typedef void* notify_handle_t;
 
 void notify_initialize();
-notify_handle_t notify(const char* msg, const char* body, const char* icon);
-void notify_update(notify_handle_t h, const char* msg, const char* body, const char* icon);
+notify_handle_t notify(const char* msg, const char* body, const char* icon, gint value);
+void notify_update(notify_handle_t h, const char* msg, const char* body, const char* icon, gint value);
 
 #endif /* PASYSTRAY_NOTIFY_H */
