@@ -1,7 +1,7 @@
 /***
   This file is part of PaSystray
 
-  Copyright (C) 2011, 2012 Christoph Gysin
+  Copyright (C) 2011-2015  Christoph Gysin
 
   PaSystray is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
@@ -31,17 +31,13 @@ pa_context* context = NULL;
 
 static pa_glib_mainloop* m = NULL;
 static pa_proplist* context_proplist = NULL;
-static char* server = NULL;
 
 void pulseaudio_init(menu_infos_t* mis)
 {
     if(!(m = pa_glib_mainloop_new(g_main_context_default())))
         pulseaudio_quit("pa_glib_mainloop_new() failed.");
 
-    pulseaudio_prepare_context();
-    pa_context_set_state_callback(context, pulseaudio_context_state_cb, mis);
-
-    pulseaudio_connect();
+    pulseaudio_connect(mis);
 }
 
 void pulseaudio_destroy()
@@ -49,7 +45,7 @@ void pulseaudio_destroy()
     pulseaudio_quit(NULL);
 }
 
-void pulseaudio_prepare_context()
+void pulseaudio_prepare_context(menu_infos_t* mis)
 {
     pa_mainloop_api* mainloop_api = pa_glib_mainloop_get_api(m);
 
@@ -59,17 +55,41 @@ void pulseaudio_prepare_context()
     context = pa_context_new_with_proplist(mainloop_api, NULL, context_proplist);
     if(!context)
         pulseaudio_quit("pa_context_new() failed.");
+
+    pa_context_set_state_callback(context, pulseaudio_context_state_cb, mis);
 }
 
-void pulseaudio_connect()
+void pulseaudio_connect(menu_infos_t* mis)
 {
-    if(pa_context_connect(context, server, 0, NULL) < 0)
+    pulseaudio_prepare_context(mis);
+
+    if(pa_context_connect(context, NULL, PA_CONTEXT_NOFAIL, NULL) < 0)
     {
-        g_message("pa_context_connect() failed: ");
-        pulseaudio_quit(pa_strerror(pa_context_errno(context)));
+        g_warning("pa_context_connect() failed: %s",
+            pa_strerror(pa_context_errno(context)));
     }
 }
 
+void pulseaudio_reconnect_cb(pa_mainloop_api *api, pa_time_event *event, const struct timeval *tv, void *userdata)
+{
+    menu_infos_t* mis = userdata;
+    pulseaudio_connect(mis);
+}
+
+void pulseaudio_reconnect(menu_infos_t* mis)
+{
+    systray_impl_set_icon(mis->systray, "pasystray");
+
+    pa_mainloop_api* api = pa_glib_mainloop_get_api(m);
+
+    struct timeval tv;
+    struct timeval* delay_tv = pa_gettimeofday(&tv);
+    pa_timeval_add(delay_tv, 1 * PA_USEC_PER_SEC);
+
+    // reconnect with 1s delay
+    api->time_new(api, delay_tv, pulseaudio_reconnect_cb, mis);
+}
+
 void pulseaudio_context_state_cb(pa_context* c, void* userdata)
 {
     menu_infos_t* mis = userdata;
@@ -77,7 +97,7 @@ void pulseaudio_context_state_cb(pa_context* c, void* userdata)
     switch(pa_context_get_state(c))
     {
         case PA_CONTEXT_UNCONNECTED:
-            g_message("PulseAudio context unconnected!");
+            g_warning("PulseAudio context unconnected!");
             break;
 
         case PA_CONTEXT_READY:
@@ -97,24 +117,40 @@ void pulseaudio_context_state_cb(pa_context* c, void* userdata)
                         PA_SUBSCRIPTION_MASK_SOURCE |
                         PA_SUBSCRIPTION_MASK_SINK_INPUT |
                         PA_SUBSCRIPTION_MASK_SOURCE_OUTPUT |
+                        PA_SUBSCRIPTION_MASK_MODULE |
                         PA_SUBSCRIPTION_MASK_SERVER,
                         pulseaudio_subscribed_cb, mis));
             break;
         }
 
         case PA_CONTEXT_FAILED:
-        case PA_CONTEXT_TERMINATED:
-            g_message("PulseAudio terminated!");
+            g_warning("[pulseaudio] context failed!");
             menu_infos_clear(mis);
             pa_context_unref(context);
-            pulseaudio_prepare_context();
-            g_message("reconnecting...");
-            pulseaudio_connect();
+
+            g_debug("[pulseaudio] trying again...");
+            pulseaudio_reconnect(mis);
+            break;
+
+        case PA_CONTEXT_TERMINATED:
+            g_warning("[pulseaudio] context terminated!");
+            menu_infos_clear(mis);
+            pa_context_unref(context);
+
+            g_debug("[pulseaudio] reconnecting...");
+            pulseaudio_reconnect(mis);
             break;
 
         case PA_CONTEXT_CONNECTING:
+            g_debug("[pulseaudio] connecting...");
+            break;
+
         case PA_CONTEXT_AUTHORIZING:
+            g_debug("[pulseaudio] authorizing...");
+            break;
+
         case PA_CONTEXT_SETTING_NAME:
+            g_debug("[pulseaudio] setting name...");
             break;
     }
 }
@@ -132,13 +168,13 @@ void pulseaudio_subscribed_cb(pa_context* c, int success, void* userdata)
                 pulseaudio_sink_input_init_cb, &mis->menu_info[MENU_INPUT]));
     pa_operation_unref(pa_context_get_source_output_info_list(context,
                 pulseaudio_source_output_init_cb, &mis->menu_info[MENU_OUTPUT]));
+    pa_operation_unref(pa_context_get_module_info_list(context,
+                pulseaudio_module_init_cb, &mis->menu_info[MENU_MODULE]));
 }
 
 void pulseaudio_event_cb(pa_context* c, pa_subscription_event_type_t t, uint32_t index, void* userdata)
 {
-#ifdef DEBUG
     pulseaudio_print_event(t, index);
-#endif
 
     menu_infos_t* mis = userdata;
 
@@ -185,6 +221,10 @@ void pulseaudio_event_new(pa_subscription_event_type_t facility, uint32_t index,
             pa_operation_unref(pa_context_get_source_output_info(context, index,
                         pulseaudio_source_output_add_cb, &mis->menu_info[MENU_OUTPUT]));
             break;
+        case PA_SUBSCRIPTION_EVENT_MODULE:
+            pa_operation_unref(pa_context_get_module_info(context, index,
+                        pulseaudio_module_add_cb, &mis->menu_info[MENU_MODULE]));
+            break;
         default:
             break;
     }
@@ -214,6 +254,10 @@ void pulseaudio_event_change(pa_subscription_event_type_t facility, uint32_t ind
             pa_operation_unref(pa_context_get_source_output_info(context, index,
                         pulseaudio_source_output_change_cb, &mis->menu_info[MENU_OUTPUT]));
             break;
+        case PA_SUBSCRIPTION_EVENT_MODULE:
+            pa_operation_unref(pa_context_get_module_info(context, index,
+                        pulseaudio_module_change_cb, &mis->menu_info[MENU_MODULE]));
+            break;
         default:
             break;
     }
@@ -238,18 +282,20 @@ void pulseaudio_event_remove(pa_subscription_event_type_t facility, uint32_t ind
         case PA_SUBSCRIPTION_EVENT_SOURCE_OUTPUT:
             menu_info_item_remove(&mis->menu_info[MENU_OUTPUT], index);
             break;
+        case PA_SUBSCRIPTION_EVENT_MODULE:
+            menu_info_item_remove(&mis->menu_info[MENU_MODULE], index);
+            break;
         default:
             break;
     }
 }
 
-#ifdef DEBUG
 void pulseaudio_print_event(pa_subscription_event_type_t t, uint32_t index)
 {
     pa_subscription_event_type_t type = t & PA_SUBSCRIPTION_EVENT_TYPE_MASK;
     pa_subscription_event_type_t facility = t & PA_SUBSCRIPTION_EVENT_FACILITY_MASK;
 
-    g_message("event %s on %s (%u)",
+    g_debug("event %s on %s (%u)",
         (type == PA_SUBSCRIPTION_EVENT_NEW) ? "new" :
         (type == PA_SUBSCRIPTION_EVENT_CHANGE) ? "change" :
         (type == PA_SUBSCRIPTION_EVENT_REMOVE) ? "remove" :
@@ -258,11 +304,11 @@ void pulseaudio_print_event(pa_subscription_event_type_t t, uint32_t index)
         (facility == PA_SUBSCRIPTION_EVENT_SOURCE) ? "source" :
         (facility == PA_SUBSCRIPTION_EVENT_SINK_INPUT) ? "sink-input" :
         (facility == PA_SUBSCRIPTION_EVENT_SOURCE_OUTPUT) ? "source-output" :
+        (facility == PA_SUBSCRIPTION_EVENT_MODULE) ? "module" :
         (facility == PA_SUBSCRIPTION_EVENT_SERVER) ? "server" :
         "unknown",
         index);
 }
-#endif
 
 void pulseaudio_server_init_cb(pa_context* c, const pa_server_info* i, void* userdata)
 {
@@ -329,14 +375,16 @@ void pulseaudio_sink_add(const pa_sink_info* i, int is_last, void* userdata, gbo
     if(is_last)
         return;
 
-    if(is_new)
+    menu_info_t* mi = userdata;
+    menu_infos_t* mis = mi->menu_infos;
+
+    if(is_new && mis->settings.notify != NOTIFY_NEVER)
     {
         gchar* msg = g_strdup_printf("new sink \"%s\"", i->description);
-        notify(msg, i->name, NULL);
+        notify(msg, i->name, NULL, -1);
         g_free(msg);
     }
 
-    menu_info_t* mi = userdata;
     char* tooltip = sink_info_str(i);
     menu_info_item_update(mi, i->index, i->name, i->description, &i->volume,
             i->mute, tooltip, NULL, NULL, -1);
@@ -371,18 +419,20 @@ void pulseaudio_source_add(const pa_source_info* i, int is_last, void* userdata,
 
     const char* class = pa_proplist_gets(i->proplist, PA_PROP_DEVICE_CLASS);
 
+    menu_info_t* mi = userdata;
+    menu_infos_t* mis = mi->menu_infos;
+
     // ignore monitors
-    if(class && g_str_equal(class, "monitor"))
+    if(!mis->settings.monitors && class && g_str_equal(class, "monitor"))
         return;
 
-    if(is_new)
+    if(is_new && mis->settings.notify != NOTIFY_NEVER)
     {
         gchar* msg = g_strdup_printf("new source \"%s\"", i->description);
-        notify(msg, i->name, NULL);
+        notify(msg, i->name, NULL, -1);
         g_free(msg);
     }
 
-    menu_info_t* mi = userdata;
     char* tooltip = source_info_str(i);
     menu_info_item_update(mi, i->index, i->name, i->description, &i->volume,
             i->mute, tooltip, NULL, NULL, -1);
@@ -471,6 +521,39 @@ void pulseaudio_source_output_add(const pa_source_output_info* i, int is_last, v
     g_free(tooltip);
 }
 
+void pulseaudio_module_init_cb(pa_context* c, const pa_module_info* i, int is_last, void* userdata)
+{
+    pulseaudio_module_add(i, is_last, userdata, FALSE);
+}
+
+void pulseaudio_module_add_cb(pa_context* c, const pa_module_info* i, int is_last, void* userdata)
+{
+    pulseaudio_module_add(i, is_last, userdata, TRUE);
+}
+
+void pulseaudio_module_change_cb(pa_context* c, const pa_module_info* i, int is_last, void* userdata)
+{
+    pulseaudio_module_add(i, is_last, userdata, FALSE);
+}
+
+void pulseaudio_module_add(const pa_module_info* i, int is_last, void* userdata, gboolean is_new)
+{
+    if(is_last < 0)
+    {
+        g_message("Failed to get source information: %s", pa_strerror(pa_context_errno(context)));
+        return;
+    }
+
+    if(is_last)
+        return;
+
+    menu_info_t* mi = userdata;
+    char* tooltip = module_info_str(i);
+    menu_info_item_update(mi, i->index, NULL, i->name, NULL, 0, tooltip, NULL,
+            NULL, -1);
+    g_free(tooltip);
+}
+
 void pulseaudio_quit(const char* msg)
 {
     if(msg)
